//-----------------------------------------------------------------------//
//  xdata.h                                                              //
//-----------------------------------------------------------------------//

#ifndef XDATA
#define XDATA 1

#ifndef LITTLE_ENDIAN
#define LITTLE_ENDIAN
#endif

#include <config.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#ifdef LINUX
#include <Xlocale.h>
#else
#include <locale.h>
#endif
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h> 
#include <sys/times.h>
#include "complex.h"   // Use our own complex.h to avoid the bugs
                       // in the g++ one.

                       // Needed for semaphores & ipc
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <Xm/XmAll.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>         // For definition of XA_WINDOW  
#include <X11/Xresource.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <Xbae/Matrix.h>       // From ftp.x.org/contrib
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include "xdataa.h"
#include <gmp.h>

//---------------#define's-------------------------//

#define min(a,b) (((a)<(b)) ? (a) : (b))
#define max(a,b) (((a)>(b)) ? (a) : (b))
#define swap(a,b) { g.swap_temp=(a); (a)=(b); (b)=g.swap_temp; }  
#define fswap(a,b) { g.fswap_temp=(a); (a)=(b); (b)=g.fswap_temp; }  
#define dswap(a,b) { g.dswap_temp=(a); (a)=(b); (b)=g.dswap_temp; } 
#define sgn(a) ( (a)>0 ? (1) : ( (a)<0 ? (-1) : (0) ) )
#define fsgn(a) ( (a)>0.0 ? (1.0) : ( (a)<0.0 ? (-1.0) : (0.0) ) )
#define square(a) ((a)*(a))
#define uchar unsigned char
#define uint unsigned int
#define ulong unsigned long
#define ushort unsigned short int
#define MAXWINDOWS 10
#ifndef MAXINT
 #define MAXINT 2147483647
#endif
#define PI 3.141592653589793238462643

#define XmACB XmAnyCallbackStruct
#define XmSBCB XmSelectionBoxCallbackStruct
#define XmFSBCB XmFileSelectionBoxCallbackStruct
#define XtEH XtEventHandler
#define XtP XtPointer
#define XtCBP XtCallbackProc
#define XBMECB XbaeMatrixEnterCellCallbackStruct
#define XBMSCCB XbaeMatrixSelectCellCallbackStruct
#define XBMWCCB XbaeMatrixWriteCellCallbackStruct 
#define XBMMVCB XbaeMatrixModifyVerifyCallbackStruct 
#define XBMLACB XbaeMatrixLabelActivateCallbackStruct 
#define XBMLCCB XbaeMatrixLeaveCellCallbackStruct 
#define XBMDCCB XbaeMatrixDrawCellCallbackStruct 
#define XBMTCCB XbaeMatrixTraverseCellCallbackStruct 

//--------------constants-------------------------//

const int NOOFMENUS=9;          // no.of menus across top
const int MACROLENGTH=65536;    // max.no.of chars in macro
const int FILENAMELENGTH=1024;  // max. length of a filename

////  Write string
const int PERM=1;
const int TEMP=0;
const int SAME=0;

////  Error codes
const int OK=0;
const int NOMEM=1;
const int ABORT=2;
const int NOTFOUND=3;
const int ERROR=4;
const int NOIMAGES=5;
const int GOTNEW=6;
const int UNKNOWN=7;
const int CANTCREATEFILE=8;
const int CRITERR=9;
const int ZEROLENGTH=10;
const int QUIT=11;
const int BADPARAMETERS=12;
const int YES=13;
const int NO=14;
const int CANCEL=15;

////  For windows
const int CLOSE=0;
const int OPEN=1;
const int HIDE=2;
const int UNHIDE=3;

////  Movement & positioning modes
const int OFF=0;
const int ON=1;
const int UP=1;
const int DOWN=2;
const int LEFT=3;
const int RIGHT=4;
const int START=5;
const int CONTINUE=6;
const int END=7;
const int CENTER=8;

////  Math functions
const int ADD=0;
const int SUBTRACT=1;
const int MULTIPLY=3;
const int DIVIDE=4;
const int NEGATIVETOZERO=5;
const int LOG10=6;
const int LOG=7;
const int RAMP=8;
const int EXPONENTIATE=9;
const int SQRT=10;
const int CBRT=11;
const int EXPONENTIAL=12;
const int TENTOX=13;
const int TWOTOX=14;
const int RANDOM=15;
const int INVERSE=16;
const int SIN=17;
const int COS=18;
const int TAN=19;
const int MIN=20;
const int MAX=21;
const int INTEGRATE=22;
const int DIFFERENTIATE=23;
const int SETTOSTRING=24;
const int SETTONUMBER=25;
const int EQU=26;
const int INT=27;
const int INT2=27;
const int LN=28;
const int MOD=29;
const int MODINVERSE=30;
const int MODPWR=31;

//// math functions for hex numbers
const int AND=29;
const int OR=30;
const int XOR=31;
const int NOT=32;
const int CONVERTTOHEX=33;
const int CONVERTTODEC=34;

////  Global states
const int RESET=-2;
const int ERASE=-1;
const int NORMAL=0;
const int MOVE=3;
const int COPY=4;
const int POSITIONBOX=5;
const int DENSITOMETRY=6;
const int GETBOX=7;
const int REDRAW=8;
const int BUSY=9;
const int INITIALIZING=10;
const int INMENU=11;
const int GETNEXTPOINT=12;
const int CALCULATING=13;
const int CURVE=14;
const int SELECT=15;
const int GETPOINT=16;
const int GETLINE=17;
const int DRAWING=18;
const int MOVEIMAGE=19;
const int DENSITOMETRY2=20;
const int DENSITOMETRY3=21;
const int DENSITOMETRY4=22;
const int READING=23;
const int ROTATEPALETTE=24;
const int WAITCHAR=25;
const int FILL=26;
const int WARP=27;
const int CALIBRATE=28;
const int PRINTING=29;
const int ALTKEY=30;
const int TESTING=31;
const int ERASING=32;
const int NEWIMAGE=33;
const int GARBAGE_COLLECT=34;
const int MACRO=35;
const int SCANNING=36;
const int XTWARNING=37;
const int PASTE=38;
const int MESSAGE=40;
const int ACQUIRE=41;
const int MESSAGE_INVERTED=42;
const int PAUSE=43;
const int EXPOSE=44;

////  Number types 
const int INTEGER=0;
const int FLOAT=1;
const int INTARRAY=2;
const int DOUBLE=4;
const int CHARACTER=4;
const int WORD=5;
const int HEXADECIMAL=6;
const int RGBVALUE=7;
const int RGBHEX=8;

////  Number types f/data entry
/*const int NORMAL=0; elswhere*/
const int DEC=0;
const int HEX=1;
const int OCT=2;
const int BIN=3;


////  List types in dialog boxes
const int RADIO=0;
const int RADIOTEXT=1;
const int RADIOVALUE=2;
const int RADIOVALUEWIDE=3;
const int RADIOPUSHBUTTON=4;


////  Box types in dialog boxes
const int TOGGLE=0;
const int INTCLICKBOX=1;
const int FILENAME=2;
const int MULTICLICKBOX=3;
const int NON_EDIT_LIST=4;
const int STRING=5;
const int LIST=6;
const int LABEL=7;
const int NO_LABEL_LIST=8;
const int DOUBLECLICKBOX=9;
const int MULTIFILENAME=10;
const int TOGGLESTRING=11;        // Toggle and get a string
const int TOGGLEMULTICLICKBOX=12; // Toggle and get RGB value
const int SCROLLEDSTRING=13;      // Get string using scrolled text widget
const int PUSHBUTTON=14;       

////  Message types
const int NONE=0;
const int INFO=1;
const int PROMPT=2;
const int QUESTION=3;
// ERROR=4
const int WARNING=5;
const int WORKING=6;
const int DONE=7;
const int YESNOQUESTION=8;

////  List add item mode
const int BOTTOM=0;
const int TOP=1;

////  Filename mode
const int SINGLE=1;
const int MULTIPLE=2;

const int CELLWIDTH = 12;      // Width of each cell to display
// Dont make this too big or it will be slow startup
const int CELLBYTES = 256;     // Size of string in each cell
const int EDITORSIZE = 65536;  // Why would anyone ever need more than this

////  Digit types
const int DIGIT=1;
const int POINT=2;
const int SIGN=3;
const int EXPONENT=4;
const int HEXDIGIT=5;

////  Dif headers
const int  TABLE=1;
const int  VECTORS=2;
const int  TUPLES=3;
/*const int  ERROR=4;*/
const int  COMMENT=5;
const int  SIZE=6;
/*const int  LABEL=7;*/
const int  PERIODICITY=8;
const int  MINORSTART=9;
const int  TRUELENGTH=10;
const int  DISPLAYUNITS=11;
const int  BOT=12;
const int  EOD=13;
const int  V=14;
const int  NA=15;
const int  DATA=16;
const int  TRUE_VALUE=17;
const int  FALSE_VALUE=18;

const int FACT=0;
/* PERM is already 1*/
const int COMB=2;

//-----------Typedefs------------------------------//

typedef struct PaletteCallbackStruct
{  int x;
   int y;
   int focus;
} pcb_struct;

typedef struct SElectCountStruct
{    int count;
     int *start;
     int *end;
} SelectCountStruct;

typedef struct SElectStruct
{    double *x;            //// value if it's a number
     uchar *isnumber;      //// array of whether each one is a number or string
     char **s;             //// string if it's a string
     int noofcols;         //// returns no. of columns in case we need to paste  
} SelectStruct;

typedef struct Clickboxinfo 
{     char *title;   
      char *filename;   
      char *dirmask;
      char *path;
      char *buttonlabel;   // For BUTTONVALSLIDER button label
      int answer;          // Generic answer for single clickbox
      double  fanswer;      // Generic answer for single clickbox
      int *answers;        // Generic array of answers for multiclickbox. Must
                           //   be a ptr so all individual clickboxes can see
                           //   each other's data. 
      int k;               // Index of which slider in multiclickbox 
      int button;          // Generic button no.
      int startval;
      double  fstartval;
      int minval;
      int maxval;
      double  fminval;
      double  fmaxval;
      int decimalpoints;   // No. of decimal points to display ints as floats
      int type;            // Data type desired (INTEGER or FLOAT)
      int wantpreview;
      int allowedit;       // 0=cant change 1=ok to change 2=each line is command
      int ask;             // 0=dont ask user (eg for filename) 1=ask user
      int editrow;
      int editcol;
      int position;
      int maxstringsize;   // Maximum length of a string to edit
      Widget form;         // The parent form widget that called clickbox
      Widget listwidget;   // The child list widget of the clickbox if any
      int count;           // No. of items in list if any
      Widget widget[20];   // Generic list of Widgets to be changed by callback
      Widget *w;           // List of Widgets created by callback, to be destroyed
      int wc;              // No. of Widgets in list w to destroy
      char **list;         // List of strings 
      int listsize;        // No. of elements in list
      int helptopic;       // Help topic number
      void *ptr;           // Generic pointer
      int done;            // Flag indicating done
      int wantdragcb;      // 1=f1,f2,f3 called while dragging mouse 0=called only
                           //    when value changes. Set to 0 if f1-f3 are slow.
                           //    (used for multiclickbox only)
      void(*f1)(int answer);       // Function to call with integer arg
      void(*f2)(double  answer);    // Function to call with double  arg
      void(*f3)(int answers[10]);  // Function to call with integer array arg
                           // Function to call with integer and clickboxinfo*
      void(*f4)(int answer, Clickboxinfo *c);  
      int callbacktype;    //0=no callback 1=editclosecb
} clickboxinfo;


typedef struct Listinfo
{     char* title;         // Title of list box
      char** info;         // Contents of list (strings)
      int size;            // No. of items
      int itemstoshow;     // No. of items to show at a time
      int firstitem;       // 1st item to show
      int wantsort;        // 1=sort items before showing list
      int helptopic;       // Help topic 
      int *selection;      // Return for answer selected by user. This must
                           //   point to an actual variable, even if no selection
                           //   is wanted. (e.g.,  list.selection = &crud; )
      int allowedit;       // 1=allow user to change items in list
                           // The following items don't need to be specified if
                           //   allowedit is 0.
      char* edittitle;     // Title of menu box if user edits a line *
      int editcol;         // Left most column user is allowed to edit
      int editrow;         // 1st row that can be edited
      int width;           // Minimum width in pixels of dialog widget to create 
                           //  (If 0, width is calculated from 'info').
      Widget form;         // Form widget containing list
      Widget widget;       // Generic widget
      Widget browser;      // Browser widget
      int transient;       // 1=list pops up  0=user must position the list
      int maxstringsize;   // Maximum allowable size for an 'info[]' item
      int wantfunction;    // 1=call function f1
      void (*f1)(int answer);    // Function to call when click on list item
      void (*f2)(int listno);    // Function to call to update list (only used if 
                           //    autoupdate is 1)
      int autoupdate;      // 1=automatically update list after every command        
} listinfo;


const int DCOUNT=30;                 // No. of box or menu items
const int RCOUNT=30;                 // No. of radio items
typedef struct dialoginfo
{                                /* Select only 1 from each set             */
    char title[50];              /* Title across top                        */ 
    char radio[RCOUNT][DCOUNT][100]; /* 10 sets of up to DCOUNT options length 100 */

    int radiono[RCOUNT];             /* no.of actual radio groups (max=10)      */
    int radiotype[RCOUNT][DCOUNT];   /* Input method for obtaining answer in a  */
                                 /* radio box (0=no answer, 1=message(),    */
                                 /* 2, clickbox().                          */

                                 /* Answers returnable from a radio box     */
                                 /* radioanswer[radioboxno][item_no]        */
                                 /* Maximum 100 characters                  */
    char radioanswer[RCOUNT][DCOUNT][100];
                                 
    int radiomin[RCOUNT][DCOUNT];    /* Minimum int value returnable in radio   */
    int radiomax[RCOUNT][DCOUNT];    /* Maximum int value returnable in radio   */
    int radiocomeback[RCOUNT];       /* 1=need to return if radio item changes  */
    int radioset[DCOUNT];        /* Radio box selection number              */

    int boxtype[DCOUNT];         /* Character type to be returned in answer */
                                 /* box (0=string, 1=integer). If integer it*/
                                 /* uses clickbox to select the integer.    */
                                    
    int boxmin[DCOUNT];          /* Minimum value returnable in box         */
    int boxmax[DCOUNT];          /* Maximum value returnable in box         */
    int wantfunc[DCOUNT];        /* Flag if clickbox in answer box calls    */
                                 /*   a callback                            */ 
    
                                 /* For selecting parameters in dialog box  */
                                 /* 1 set of up to DCOUNT multiple choices  */
    char boxes[DCOUNT][100];     /* or up to DCOUNT strings in boxes        */
    int boxset[DCOUNT];          /* Toggle box setting (0 or 1)             */
    int boxcount[DCOUNT];        /* Returns number of items in boxlist      */
 
    int noofradios;              /* no. of radio groups                     */
    int noofboxes;               /* no. of string boxes + check boxes       */

                                 /* Prior value of most recently-changed    */
                                 /* string if non-zero                      */
    char lastanswer[DCOUNT][100];
                                 /* Corresponding strings returned from box */
                                 /* This has to be >= 2 chars longer than   */
                                 /*  the maxlength used in 'message'.       */
    char answer[DCOUNT][FILENAMELENGTH]; 
    char path[FILENAMELENGTH];   /* Path to start looking for filename      */
    int helptopic;               /* ID # of associated help topic           */
    char **boxmenu[DCOUNT];      /* Menu items for box (dynamically alloc'd)*/
                                 /*   max of DCOUNT menu items/listbox      */
    char *boxmenutitle[DCOUNT];  /* Menu title                              */
    char **boxlist[DCOUNT];      /* Multi-filename list for box             */
    int boxmenuselection[DCOUNT];/* Number of menu item selected.           */
                                 /*  boxmax[k] is the total no. of items.   */
    int boxcomeback[DCOUNT];     /* 1=Need to return if box item changes    */
    int startcol[DCOUNT];        /* Leftmost allowable editing position     */
                                 /*  (in strings, not lists)                */
    int needrepeat;              /* 1=Comeback was hit, need to call dialog */
                                 /*   again                                 */
    char *message1;              /* Error message to display first          */
    char *message2;              /* Error message to display first          */
    int messagemode;             /* Message mode of error message           */
    double  clickboxstep[DCOUNT];     /* Step distance for slider in clickbox    */
    Widget radiowidget[RCOUNT][DCOUNT];  /* Radio button Widget (needed in callback)*/ 
    Widget radioslider[RCOUNT][DCOUNT];  /* Radio button Widget (needed in callback)*/ 
                                     /* Answer box Widget (needed in callback)  */
    Widget boxwidget[DCOUNT][4];     /* Max of 4 widgets per line (eg, r,g,b or */
                                     /* a toggle button + string).              */
                                
    int transient;               /* Whether to make window transient        */
    int bpp;                     /* Bits/pixel to use in multiclickbox      */
    int want_changecicb;         /* 1=call dialogchangecicb when changing ci*/
    int keep;                    /* ON = dialog stays on screen             */
                                 /* OFF = dialog automatically unmanaged    */
                                 /* RESET = unmanage a dialog               */
    clickboxinfo *c[DCOUNT];     /* clickboxinfo struct needed in callbacks */             
                                       
                                 /* The following are needed in case dialog */
                                 /*    stays on the screen                  */
    Widget *widget;              /* Array of widgets to delete              */
    Widget form;                 /* Form widget to delete                   */
    int widgetcount;             /* no. of widgets on form                  */
    clickboxinfo *clickboxdata;  /* Array of clickboxinfo structs           */

                                 /* Functions to call when any button is clicked */
    void (*f1)(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);   
    void (*f2)(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);   
    void (*f3)(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);   
    void (*f4)(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);   


    XtCBP radiocb[10];           /* Callback for RADIOPUSHBUTTONs           */
                                 /* Labels for RADIOPUSHBUTTONs             */
    char radiopushbuttonlabel[RCOUNT][100];
    void *radiopushbuttonptr[RCOUNT];/* Data to send to RADIOPUSHBUTTONs        */
    int list_editrow[DCOUNT];    /* Minimum list item that can be edited    */
    int list_editcol[DCOUNT];    /* Minimum column in list item one can edit*/
                                 /* Max column in list item                 */
    int list_maxstringsize[DCOUNT];

} Dialog;


typedef struct Pushbuttoncbstruct
{  int button;
   Widget widget;
   char string[100];
   char command[100];
} pushbuttoncbstruct;

typedef struct BUttonStruct
{
    char *label;
    char *cmd;
} ButtonStruct;


typedef struct SPreadsheet
{
    Widget   spreadsheet_widget;       // Xbae widget
    Window   spreadsheet_form_window;  // form widget (parent of Xbae widget)
    char ***text;                      // User-entered text for each cell
    char ***disp;                      // Text to display for each cell
    char ***form;                      // Text to display for each cell
    char **title;                      // Title for each column, user settable
    char **otitle;                     // Backup copy of column title
    char **rowlabel;                   // Title for each row, user settable
    char **collabel;                   // Title for each column (1, 2, ...) fixed
    double **x;                        // Actual numbers calculated for each cell
    double  *x_1d;                     // 
    uchar  **selected;                 // 1=cell is selected 0=not selected
    uchar   *selected_1d;              //
    int     *maxrow;                   // Highest non-blank row in each column 
    int      maxcol;                   // Highest non-blank column
    uchar  **number;                   // 1=cell contains a number 0=cell is undetermined
    uchar   *number_1d;                //
    uchar  **string;                   // 1=cell contains a string 0=cell is undetermined
    uchar   *string_1d;                //
    uchar  **formula;                  // 1=cell contains a formula
    uchar   *formula_1d;               //
    uchar  **hex;                      // 1=cell contains a hex number
    uchar   *hex_1d;                   //
    uchar   *column_selected;          // 1=column was selected
    uchar   *row_selected;             // 1=row was selected
    int      current_row;              // Cursor row
    int      current_col;              // Cursor column
    int      rows;                     // Total no. of rows
    int      cols;                     // Total no. of cols    
    int      titlecols;                // Total no. of column labels
    int      hashtitlecols;            // Total no. of column titles specified with #
    int      titlerows;                // Total no. of row labels
    int      header_exists;            // 1=spreadsheet has a header to save
} Spreadsheet;


typedef struct GLobals
{
    ButtonStruct **button;     // User-defined buttons at left
    Display  *display;         // Connection to X display  
    GC       gc;               // The graphics context for main
    uint     mask;
    Widget   main_widget;
    Widget   dialogdone;
    Widget   menu;
    Widget   menubar;
    Window   rootwindow;
    Window   mainwindow;
    Visual  *visual;
    XFontStruct  *fontstruct;  // Default font
    XmFontList    fontlist;    // Font list for Motif 1.2 and below
    Widget menupane[30];
    int swap_temp;
    double   fswap_temp;       // Temporary double  for swap macro
    double   dswap_temp;       // Temporary double  for swap macro
    int      rr, cc;           // Column and row passed to expression evaluator
    char    *appname;          // Application's name
    char     homedir[FILENAMELENGTH];        // $HOME/.xdata directory for xdata.ini
    char    *commands;  
    char     currentdir[FILENAMELENGTH];       // Current directory
    char     filename[FILENAMELENGTH];         // Filename
    char     gvfilename[FILENAMELENGTH];       // Filename for gv
    char    *font;             // Name of font
    char    *macrotext;
    char     printer_command[FILENAMELENGTH];  
    char     printer_filename[FILENAMELENGTH]; 
    char     startdir[FILENAMELENGTH];         // Start-up directory
    char     waveletpath[FILENAMELENGTH];      // Where to look for wavelets
    char     version[100];

    int      main_xsize;
    int      main_ysize;
    int      bcolor;
    int      fcolor;
    int      scolor;              // string color
    int      ecolor;              // equation color
    int      hexcolor;            // hex number color
    int      cols; 
    int      getout;              // Flag for 'cancel'
    int      key_ascii;
    int      key_alt;
    int      key_ctrl;
    int      key_shift;
    int      nbuttons;
    int      rows;
    int      screen;
    int      signif;              // Significant digits to display
    int      state;               // Program button-pressing state
    int      want_messages;
    int      xlib_cursor;
    int      want_colormaps;
    int      main_fcolor, main_bcolor, main_fixedcolor;
    int      graph_active;        // 1=files still open in graph fifo
    int      autoupdate;          // 1=automatically update graph when spreadsheet refreshed
    int      autoscale;           // 1=automatically scale graph 
    int      touched;             // 1=data has been changed
    int      string;              // 1=string 0=number
    int      hexmode;             // 0=normal 1=hex for spreadsheet
    int      message_count;       // Number of messages shown since last cleared
    int      busy;
    int      number_type;         // dec, hex, octal, bin for spreadsheet
} Globals;

typedef struct WAvelet
{   char *filename;
    double **filter;
    double **rfilter;
    int size;
    int ioffset;
    int joffset;
} Wavelet;


 
 
//--------------Function Prototypes------------------//

Widget addlabel(Widget parent, char *string, int position, double  x1, double  y1, 
       double  x2, double  y2);
Widget createmenu(Widget parent, char *name, char *label, char mnemonic);
Widget createmenuitem(Widget parent, char *name, char *label,
       char mnemonic, XtCallbackProc callback, const char *cbArg);
Widget createmenuseparator(Widget parent, char *name);
Widget openwindow(listinfo *l, Widget *listwidget, Widget *w, int *count);

char  *basefilename(char* pathname);
#ifndef HAVE_GCVT 
char  *gcvt(double  val, int signif, char *buf);
#endif
char  *getfilename(char *oldtitle);
char **getfilenames(int *count);
char  *itoa(int value, char *str, int radix);
char  *ltoa(long int value, char *str, int radix);
char  *strlwr(char *s);
char  *strupr(char *s);

uchar **make_2d_alias(uchar *b, int x, int y);

int addmainbutton(Widget parent, char *name, char *command, XtCallbackProc cb);
void add_variable(char *name, double  value);
void areaundercurve(void);
double  read_variable(char *name);
int between(int a,int b,int c);
int between(double a, double b, double c);
int cant_overwrite_file(char* filename);
int cint(double a);
int clickbox(char* title, int startval, int minval, int maxval, 
      void f(int answer), int helptopic);
int clickbox(char* title, int startval, int minval, int maxval, void f1(int answer),
      void f2(double answer), int decimalpoints, int type, int helptopic);
int column_has_selected_numbers(int col);
int compare(const void *p, const void *q);
int count_selected_data(SelectCountStruct *sel, int wantstrings);
int count_one_dataset(int set[], int n[]);
int count_two_datasets(int set[], int n[]);
char delimiter_guess(char *filename);
int digit_type(int c);
int evaluate_conditional(char *s);
int evaluate_formula(char *formula);
int execute(char *command, char **arg, int noofargs);
int filesize(char *filename);
int get_column(SelectStruct *c, int col);
int get_selected_data(SelectStruct *c, int wantstrings, int maxn, int &noofcols);
int get_one_dataset(double *data[1], int total[1]);
int get_two_datasets(double *data[2], int total[2]);
int linreg(char **arg, int noofargs);
int linreg2(char **arg, int noofargs);
int macro(char *text);
int macro_math(int n, char **arg);
int **make_2d_alias(int *b, int x, int y);
long long int **make_2d_alias(long long int *b, int x, int y);
int message(char *heading);
int message(char *heading, int mode);
int message(char *heading, char *returnstring, int mode, int maxstringsize);
int multiclickbox(char* title, int noofboxes, clickboxinfo* a,
      void f(int values[10]), int helptopic);
int parselines(char *string, char *s, int &position);
int parse_string(char *cmd, char **buf, int *stringflag, int &noofargs, 
      int startarg, int maxlen, int maxargs, char delimiter, int skip);
int readmacrofile(char *filename, char *text);
int read_next_non_comment(FILE *fp, char *buffer);
int save_ok(int ino);
int stringtohex(char* string);
int string_index(char *text, int spacing, int start, int c);
int token_type(char *buffer, double *value1, double *value2);
int yyparse(void);

uint **make_2d_alias(uint *b, int x, int y);

double betacf(double a, double b, double x);
double betai(double a, double b, double x);
double binom(double n, double k);
double binom(int n, int k, char* tempstring);
double binprob(int n, int k, double p, int *errflag);
double binprob_gmp(ulong n, ulong k, double p, int *errflag);
double clickbox(char* title, double fstartval, double fminval,
       double fmaxval, void f(double answer), int helptopic);
double divide(double a, double b);
double expression(void);
double erf(double x);
double eval(char* string);
double eval(char* string, int col, int row);
double factor(void); 
double factorial(int n);
double factorial(double n);
double factorial(int n, char* tempstring);
double fgauss(double xx, double *a, double *dyda, int mfit);
double func(int funcID, double xx, double *a, double *dyda, int mfit);
double fvalue1(double *data0, int n0, double testval);
double fvalue2(double *data0, int n0, double *data1, int n1);
double gammln(double x);
double gammp(double a, double x);
double gammq(double a, double x);
double gmp_series_term(double a, int b, int c);
double hextonumber(char *hexstring);
double largest(int col);
double **make_2d_alias(double *b, int x, int y);
double mexp(double x, double *a, double *dyda, double mfit);
double modexp(double a, double power, double m);
double modinverse(double a, double m);
double modpwr(double a, double pwr, double m);
double other(double x, double *a, double *dyda, int mfit);
double permutations(double n, double k);
double permutations(int n, int k, char* tempstring);
double poly2(double x, double *a, double *dyda, int mfit);
double polynomial(double x, double *a, double *dyda, int mfit);
double poly6(double x, double *a, double *dyda, int mfit);
double poly7(double x, double *a, double *dyda, int mfit);
double polynomial_evaluate(int mfit, double x, double *a);
double poly2fnc(int mfit, double x, double *a);
double poly6fnc(int mfit, double x, double *a);
double poly7fnc(int mfit, double x, double *a);
double pvalue(double mean1, double sd1, int n1, double mean2, double sd2, int n2);
double notzero(double a);
double randomnumber(double dummy);
double smallest(int col);
double sum(double a, double b, double c, double d);
double sumsq(double a, double b, double c, double d);
double term(void); 
double tvalue(double mean1, double sd1, int n1, double mean2, double sd2, int n2);

void abouttheprogram(void);
void additemtolist(Widget w, Widget form, char *s, int showpos);
void addOkCancelHelpButtons(Widget form, Widget button[3], char *s1, int x1, 
   char *s2, int x2, char *s3, int x3, int *data, void *ptr);
void add_prefix(char *prefix, char *string);
void addstandardbuttons(Widget form, Widget buttons[3], int *data, void *ptr);
void add_value(double value, int row, int col);
void anova(void);
void avevar(double *data, int count, double *ave, double *var);
void baseline(double *pdata, int count, int iterations);
void bin(char **arg, int noofargs, int want_average);
void binomialprobability(char **arg, int noofargs);
void binval(size_t const size, void const * const ptr, char *tempstring2);
void changerowtitle(char **arg, int noofargs);
void changetitle(char **arg, int noofargs);
void chisq1(void);
void chisq2(void);
void chisq3(void);
void closewindow(Widget form, Widget *widgetlist, int wc);
void combinese(void);
void combinese3(void);
void convert_mpf_to_scinot_string(mpf_t aa, char *tempstring);
void convert_mpz_to_scinot_string(mpz_t aa, char *tempstring);
void converttohex(uchar *buffer, uchar *hexbuffer);
void copy(char **arg, int noofargs, int mode);
void cosvrt(double **covar, int ma, int *lista, int mfit);
void create_main_icon(void);
void create_spreadsheet(Widget parent);
void cswap(complex &a, complex &b);
void decimate(char **arg, int noofargs);
void define_user_variable(char **arg, int noofargs);
void define_variable(char *name, double value);
void delete_selected_cells(char **arg, int noofargs, int n);
void delete_clipboard(SelectStruct *c, int n);
void delete_column(char **arg, int noofargs);
void delete_row(char **arg, int noofargs);
void deselectallcells(char **arg, int noofargs, int m);
void dialogbox(dialoginfo &a);
void do_print(dialoginfo *a);
void edit(char *title, char *subtitle, char *text, int rows, int columns, 
     int maxlength, int helptopic);
void edit_cell(int row, int col);
void erase(void);
void erase_cell(int row, int col);
void errorallsets(void);
void error_message(char *s, int error);
void error_propagation(void);
void errorpropagationfrommean(void);
void errorpropagationratio(void);
void error_propagationsetvalue(void);
void errorpropagation_showresults(double mean1, double mean2, double mean3,
     double mean4, double sd1, double sd2, double sd3, double sd4);
void errorsets(void);
void estimate_values(int funcID, int mx, int my, int mz, int mp, int me, 
     int ndata, double *a, double *dyda, int mfit);
void evaluate(char **arg, int noofargs);
void evaluate_cells(void);
void event_loop(Widget w);
void exectextcb(Widget w, XtP client_data, XmACB *call_data);
void export_data(char **arg, int noofargs);
void fact(char **arg, int noofargs, int mode);
void fft(char **arg, int noofargs);
void findintegerdifferences(char **arg, int noofargs);
void findpeaks(char **arg, int noofargs);
void fix_up_spreadsheet(void);
void format_number(double xx, char *result);
void fourn(complex data[], int nn[], int ndim, int isign);
void ftest(void);
void formula_error(char *error_message);
void gaussj(double **aa, int n, int np, double **bb, int m, int mp);
void get_first_word(char *buffer, char *word, int maxlen);
void getnextchar(void);
void getword(char *buffer, char *word, char delimiter, 
    int maxlen, int &count, int &is_string);
void gcf(double *gammcf, double a, double x, double *gln);
void graph(int updating, char **arg, int noofargs);
void graph_draw(dialoginfo a);
void graph_close(void);
void gser(double *gamser, double a, double x, double *gln);
void handle_keysym(KeySym keysym, int x, int y);
void help(int topic);
void hex(char **arg, int noofargs);
void hextointeger(char *hexstring, char *decstring);
void initialize_evaluator(void);
void initialize_globals(void);
void init_xlib(void);
void insert_cell(int row, int col);
void insert_row(int row, int col);
void integertohex(char *decstring, char *hexstring);
void integraterange(char **arg, int noofargs);
void leavespreadsheet(void);
void linkedsort(char **arg, int noofargs);
void list(listinfo *l);
void lr(double *xxx, double *yyy, int n, int &i2, double &aaa, double &siga, double &bbb, double &sigb, 
        double &cc,  double &ss, double &tt, double &xj2, double &xj3, double &xk2);
void logic(void);
void long_integer_math(int y1, int y2, int y3, int noofoperands, long long int operand, int op);
void makewidgets(Widget parent);
void mannwhitney(void);
void manual(void);
void marq(char **arg, int noofargs);
void math(char **arg, int noofargs, int function);
void maximum(char **arg, int noofargs, int mode);
void move(char **arg, int noofargs);
void move_cursor(int direction);
void mrqcof(int funcID, int mx, int my, int ms, int ma, int mfit, int ndata, 
            int *lista, double *a, double **alpha, double *beta, double &chisq);
void mrqmin(int funcID, int my, int mx, int ms, int ma, int ndata, int mfit,  int nca,
     int *lista,  double &chisq, double &alamda,  double *a, double **covar, 
     double **alpha, double **da2);
void msmstotic(char **arg, int noofargs);
void myaction(Widget widget, XEvent *event, String *params, Cardinal *number_params);
void normalize(char **arg, int noofargs);
void normalize2(char **arg, int noofargs);
void normalizemax(char **arg, int noofargs);
void null(int a);
void null(int a[10]);
void null(double a);
void null(void);
void null(dialoginfo *a, int b, int c, int d, int e);
void null(void *p, int n1, int n2);
void page_down(void);
void page_up(void);
void parsecommand(char *cmd, char **arg, int &noofargs, int maxlen, int maxargs);
void parse_header(char *string, char **destination, int &noofargs, int &count, int maxlen);
void parsemacro(char *text, char *line, char *line2, int wantline, char **arg, 
   int &argno, int &opos, int &owantline, int &olineno);
void parse_title(char *string, char *destination, int &noofargs, int titlecol, int maxlen);
void path(char *pathname, char *path);
void pkc(void);
void print(char *string,int x, int y, uint fcol, uint bcol, GC *gc, Window win, 
      int wantbkg, int vertical);
void print_matrix(char *title, double **covar, int x, int y);
void print_parameters(int funcID, int mfit, int ndata, double *a, double **covar, 
    double alamda, double chisq, int usems);
void print_spreadsheet(void);
void printstatus(int state);
void pipe(void);
void poisson(char **arg, int noofargs);
void poisson2(char **arg, int noofargs);
void poisson3(char **arg, int noofargs);
void pvaluefromr(void);
void pvaluelinreg(void);
void pvaluelinregcheck(double sb1, double b1, int n, double sb2, double b2, int n2);
void range(int col, double &lowest, double &highest);
void read(char **arg, int noofargs);
void readdiffile(char *filename, char **textrow);
void readgcdfile(char *filename, char **textrow, int startrow, int startcol);
void readplainfile(char *filename, char **textrow, int startrow, 
     int startcol, char delimiter, char *stripstring, int read_what, 
     int readheader, int getcol);
void readsettings(void);
void read_wavelet(Wavelet &w);
void redefine_variable(char *name, double value);
void reduplicate(char **arg, int noofargs, int want_average);
void reformat_dif_table(char **args, int noofargs);
void reformatmassspec(char **arg, int noofargs);
void refreshspreadsheet(void);
void reload(char **arg, int noofargs);
void reorder(void);
void remove_all_cr(char *s);
void remove_prefix(int length, char *string);
void remove_spikes(char **arg, int noofargs);
void remove_leading_space(char *s1, char *s2);
void remove_trailing_characters(char *s);
void remove_trailing_space(char *s);
void replacewithmean(char **arg, int noofargs, int operation);
void resize_main_widget(int xsize, int ysize);
void reverse(char **arg, int noofargs);
void rowstats(void);
void save(char **arg, int noofargs);
void savesettings(void);
void select_cell(int row, int col, int value);
void send_event(Window w, int eventtype, int x, int y, int width, int height);
void set_background(Widget w, int row, int col);
void set_cell(int row, int col, char *text);
void set_cell_text(int row, int col, char *text);
void set_cell_value(int row, int col, double value);
void set_formula(void);
void select_column(int col, int value);
void select_down(Widget w, XEvent *e, String *s, Cardinal *n);
void select_row(int row, int value);
void select_up(Widget w, XEvent *e, String *s, Cardinal *n);
void set(char **arg, int noofargs);
void set_length(char **arg, int noofargs, int dummy);
void sethexmode(void);
void setrowtitle(int row, char *title);
void set_row_title(void);
void setcoltitle(int col, char *title);
void setallcoltitles(char **title);
void set_to_formula(void);
void set_to_string(int rr, int cc);
void set_widget_string(Widget w, char* s);
void show_coordinates(int row, int col);
void shrink(char **arg, int noofargs);
void smooth(char **arg, int noofargs);
void smooth_baseline(double * pdata, int count, int width);
void sort(char **arg, int noofargs);
void sselect(char **arg, int noofargs);
void statistics(void);
void statistics2(int set, int n, double &mean, double &sum, double &sd, double &se, double &variance);
void strip_extraneous_dashes(char *s);
void subtract_baseline(char **arg, int noofargs);
void swap_columns(char **arg, int noofargs);
void timer_callback(XtPointer client_data, XtIntervalId *timer_id);
void toggle_cell(int row, int col);
void tptest(void);
void ttest1(void);
void ttest2(void);
void ttestallvalues(void);
void ttestfrommean1(void);
void ttestfrommeanse1(void);
void ttestfrommean2(void);
void ttestfrommeanse2(void);
void ttest_unequal(void);
void warningmessagecb(String aaa);
void wavelet(char **arg, int noofargs);
void wavelet_check(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);
void wavelet_pyramidal_calc(Wavelet &wave, int direction, int ndim, int *nn, 
     int col1, int col2, int imin, int imax, double cmin, double cmax, 
     double ignmin, double ignmax);
void wavelet_transform(Wavelet &wave, double  *aa, int n, int direction);
void whatevent(int eventtype);
void whatis(char **arg, int noofargs);
void xminitializemisc(Widget parent);


//---------callbacks------------------------------//

void armboxcb(Widget w, XtP client_data, XmACB *call_data);
void cancelcb(Widget w, XtPointer client_data, XmACB *call_data);
void cancelmsgcb(Widget w, XtPointer client_data, void *call_data);
void changecb(Widget w, XtPointer client_data, XmACB *call_data);
void clickcb(Widget w, XtPointer client_data, XBMTCCB *call_data);
void clickboxcb(Widget w, XtPointer client_data, XmACB *call_data);
void columnselectcb(Widget w, XtPointer client_data, XBMLACB *call_data);
void drawcellcb(Widget w, XtPointer client_data, XBMDCCB *call_data);
void dialogclickboxcb(Widget w, XtPointer client_data, XmACB *call_data);
void dialogcursorcb(Widget w, XtP client_data, XmACB *call_data);
void dialogfilenamecb(Widget w, XtPointer client_data, XmACB *call_data);
void dialogdoubleclickboxcb(Widget w, XtPointer client_data, XmACB *call_data);
void dialoglistcb(Widget w, XtPointer client_data, XmACB *call_data);
void dialogmulticlickboxcb(Widget w, XtPointer client_data, XmACB *call_data);
void dialogokcb(Widget w, XtP client_data, XmACB *call_data);
void dialogstringcb(Widget w, XtPointer client_data, XmACB *call_data);
void dragcb(Widget w, XtPointer client_data, XBMECB *call_data);
void drawcellcb(Widget w, XtPointer client_data, XBMDCCB *call_data);
void editclosecb(Widget w, XtP client_data, XmACB *call_data);
void editcursorcb(Widget w, XtP client_data, XmACB *call_data);
void editmodifycb(Widget w, XtP client_data, XmACB *call_data);
void entercb(Widget w, XtPointer client_data, XEvent *event);
void executecb(Widget w, XtPointer client_data, void *call_data);
void disarmcb(Widget w, XtPointer client_data, XmACB *call_data);
void entercellcb(Widget w, XtPointer client_data, XBMECB *call_data);
void dragcb(Widget w, XtPointer client_data, XBMECB *call_data);
void filenamecb(Widget widget, XtPointer client_data, XmACB *call_data);
void helpcb(Widget w, XtPointer client_data, XmACB *call_data);
void fscancelcb(Widget w, XtPointer client_data, void *call_data);
void fscdcb(Widget w, XtPointer client_data, void *call_data);
void fsokcb(Widget w, XtPointer client_data, void *call_data);
void graph_update(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);   
void inputcb(Widget w, XtPointer client_data, XBMLCCB *call_data);
void leavecellcb(Widget w, XtPointer client_data, XBMLCCB *call_data);
void listcb(Widget w, XtPointer client_data, void *call_data);
void loadtextcb(Widget w, XtP client_data, XmACB *call_data);
void logic_check(Dialog *a, int radio, int radiobutton, int box, int boxbutton);
void mainbuttoncb(Widget w, XtPointer client_data, void *call_data);
void maincancelcb(Widget w, XtPointer client_data, XmACB *call_data);
void modifycb(Widget w, XtPointer client_data, XBMMVCB *call_data);
void motioncb(Widget w, XtPointer client_data, XBMMVCB *call_data);
void mousecb(Widget w, XtPointer client_data, XEvent *event);
void nomsgcb(Widget w, XtPointer client_data, void *call_data);
void okcb(Widget w, XtPointer client_data, XmACB *call_data);
void okmsgcb(Widget w, XtPointer client_data, void *call_data);
void promptmsgcb(Widget w, XtPointer client_data, void *call_data);
void pushbuttonboxcb(Widget w, XtP client_data, XmACB *call_data);
void quitcb(Widget w, XtPointer client_data, void *call_data);
void raisecb(Widget w, XtP client_data, XEvent *event);
void savecb(Widget w, XtPointer client_data, XmACB *call_data);
void savetextcb(Widget w, XtP client_data, XmACB *call_data);
void selectcb(Widget w, XtPointer client_data, XBMSCCB *call_data);
void slidercb(Widget w, XtPointer client_data, XmACB *call_data);
void swapbytes(uchar* a, int n, int size);
void togglecb(Widget w, XtPointer client_data, XmACB *call_data);
void toggleboxcb(Widget w, XtPointer client_data, XmACB *call_data);
void traversecellcb(Widget w, XtPointer client_data, XBMTCCB *call_data);
void writecellcb(Widget w, XtPointer client_data, XBMWCCB *call_data);
void printcheck(dialoginfo *a, int radio, int radiobutton, int box, int boxbutton);
void widgetraisecb(Widget w, XtP client_data, XEvent *event);

//---------end of function prototypes-------------//

#endif  // ifndef XDATA












